import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import stan
import seaborn as sb
import pickle
from my_module import *
import arviz as az
    
fin_nm = input('Data file name (*.xlsx) = ')
fout_nm = input('Output file name (*.txt) = ')
fout = open(fout_nm, 'w')
fout.write('Data file...{}\n'.format(fin_nm))
#
#       Read data from Excel file
#
xlsx_fl = pd.ExcelFile(fin_nm)
data_xlsx = pd.read_excel(xlsx_fl,  header = None)
data = data_xlsx.values
var_name = data[0]
values = data[1:]

N = len(values)
ID = []
X = np.empty(N, dtype = float)
Res = np.empty(N, dtype = int)
for t in range(N):
    ID.append(values[t][0])
    X[t] = float(values[t][1])
    Res[t] = int(values[t][2])

print('Number of data = ', N)
fout.write('\nNumber of data = {}\n\n'.format(N))

for i in range(N):
    print('{0}  {1}  {2}'.format(ID[i], X[i], Res[i]))
    fout.write('{0:>5}  {1:>10.2f}  {2:>5d}\n'.format(ID[i], X[i], Res[i]))

C1, X_mid, C2 = np.percentile(X, [20, 50, 80])
print('C1,C2 = ', C1, C2)
X_sd = (C2 - C1) / 2
print('X_sd =', X_sd, '   X_mid =', X_mid)

stan_data = {'N': N, 'X': X, 'Res': Res, 'd_mid': X_mid, 'd_sd' : X_sd}
#
#       Sampling by Stan
#
with open('PM3Cat_EqCPS3.stan', 'r') as f_stan:
    sm = stan.build(f_stan.read(), data = stan_data)
    fit = sm.sample() 

dframe = fit.to_frame()
i_data = az.from_pystan(posterior = fit, posterior_model = sm)

print(az.summary(i_data))
fout.write('\nSummary:\n')
fout.write(f'{az.summary(i_data)}\n')

az.plot_trace(i_data, var_names = ['pse', 'jnd', 'C1', 'C2'])
plt.tight_layout()
plt.show()

ck = az.plot_posterior(i_data, var_names = 'pse', point_estimate = 'median')
plt.show()
print(ck)

ck = az.plot_posterior(i_data, var_names = 'jnd', point_estimate = 'median')
plt.show()

#
#       MAP estimates for mu, sigma and C
#
#
def CalcMAPEst(samples, a = 0.05, n_points = 10000):
    """
        Calculatte a MAP estimate from a KDE graph on [Lp, Up]
        Lp and Up are 100*a/2 and 100(1-a/2) percentile points of samples
    """
    import numpy as np
    import scipy.stats as ss
    
    Lp, Up = np.percentile(samples, [100 * a/2, 100 * (1 - a/2)])  
    coord = np.linspace(Lp, Up, n_points) 
    est_pdf = ss.gaussian_kde(samples).pdf(coord)  
    map_idx = np.argmax(est_pdf) 
    MAP_Est = coord[map_idx]       
    return MAP_Est, est_pdf[map_idx]


muMAP = CalcMAPEst(dframe['pse'])[0]
sgmMAP = CalcMAPEst(dframe['sgm'])[0]
C1MAP = CalcMAPEst(dframe['C1'])[0]
C2MAP = CalcMAPEst(dframe['C2'])[0]
SgmL025 = np.percentile(dframe['sgm'], 2.5)
SgmU975 = np.percentile(dframe['sgm'], 97.5)
MuL025 = np.percentile(dframe['pse'], 2.5)
MuU975 = np.percentile(dframe['pse'], 97.5)

#
#       Prob(z < s75) = 0.75 for the standard normal distribution
#
s75 = 0.6745  
JND = sgmMAP * s75
PSE = muMAP
print('\nMAP for PSE = {0:<.2f},    95% CI for PSE = [{1:<.2f}, {2:<.2f}]'.format(
            PSE, MuL025, MuU975))
print('\nMAP for JND = {0:<.2f},    95% CI for JND= [{1:<.2f}, {2:<.2f}]'.format(
            JND, s75 * SgmL025, s75 * SgmU975))
fout.write('\nMAP for PSE = {0:<.2f},    95% CI for PSE = [{1:<.2f}, {2:<.2f}]'.format(
            PSE, MuL025, MuU975))
fout.write('\nMAP for JND = {0:<.2f},    95% CI for JND= [{1:<.2f}, {2:<.2f}]'.format(
            JND, s75 * SgmL025, s75 * SgmU975))
#
#       Data and the psychometric function
#
pf = PF(muMAP, sgmMAP)          #   Class PF from my_module.py
minX = X[np.argmin(X)]
maxX = X[np.argmax(X)]
xvalues = []
yvalues = []
yvalues_L = []
yvalues_R = []
for i in range(1001):
    t = minX + i * (maxX - minX) / 1000.0
    xvalues.append(t)
    yvalues.append(pf.v(t))
    yvalues_L.append(pf.v(t + (C2MAP-C1MAP)/2))
    yvalues_R.append(pf.v(t - (C2MAP-C1MAP)/2))
plt.title('Psychometric Functions\n' +
          ('PSE = {0:<.2f},  JND = {1:<.2f},  C = {2:<.2f}'.format
           (PSE, JND, (C2MAP-C1MAP)/2)),
          fontsize = 16)
plt.plot(xvalues, yvalues, 'k--', linewidth = 3, label = 'PF')
plt.plot(xvalues, yvalues_L, 'g-', linewidth = 3, label = 'PF(>1)')
plt.plot(xvalues, yvalues_R, 'b-', linewidth = 3, label = 'PF(>2)')
plt.yticks([0.0, 0.5, 0.75, 1.0])
#
#       Responses are scaled to values from 0 to 1, and added small fluctuations
#
Res_rn = (Res - 1) / 2 + np.random.uniform(-0.05, 0.05, len(Res))
plt.scatter(X, Res_rn, alpha = 0.5, color = 'g', label = 'Data')
plt.plot([minX, maxX], [0.0, 0.0], color = 'k')
plt.plot([minX, maxX], [1.0, 1.0], color = 'k')
plt.plot([minX, PSE, PSE], [0.5, 0.5, 0.0], color = 'k')
plt.plot([minX, PSE + JND, PSE + JND], [0.75, 0.75, 0.0], color = 'k')
plt.legend(loc = 7)
plt.xlabel('Stimulus value', fontsize = 12)
plt.ylabel('Probability', fontsize = 12)
plt.show()
#
#       Observed responses smoothed by Kernel density estimation 
#
w_1 = np.empty(len(Res))        #   Weights for Response >= 2
w_2 = np.empty(len(Res))        #   Weights for response >= 3
for t in range(len(Res)):
    w_1[t] = 0 if Res[t] < 2 else 1
    w_2[t] = 0 if Res[t] < 3 else 1

X_range = np.arange(minX, maxX, (maxX - minX) / 100)
w_k = JND / 2

while True:
    pm_1 = []           #   KDE values for Response >= 2
    pm_2 = []           #   KDE values for Response >= 3
    for v in X_range:
        pm_1.append(k_ave(v, X, w_1, w_k))
        pm_2.append(k_ave(v, X, w_2, w_k))
        
    plt.title('Observed and Estimated PFs\n' +
              ('PSE = {0:<.2f},  JND = {1:<.2f},  C = {2:<.2f}'.format(
                  PSE, JND, (C2MAP-C1MAP)/2)),
              fontsize = 16)
    plt.plot(xvalues, yvalues_L, 'g-', linewidth = 3, label = 'PF(>1)')
    plt.plot(xvalues, yvalues_R, 'b-', linewidth = 3, label = 'PF(>2)')
    plt.yticks([0.0, 0.5, 1.0])
    plt.plot(X_range, pm_1, 'g--', linewidth = 2, label = 'Res(>1)')
    plt.plot(X_range, pm_2, 'b--', linewidth = 2, label = 'Res(>2)')
    plt.xlabel('Stimulus value', fontsize = 12)
    plt.legend(loc = 'upper left')
    plt.show()
    print('\nThe current value of w_k = {}'.format(w_k))
    w_k = float(input('The nexr value of w_k = '))
    if w_k <= 0.0:
        break

fout.close()
print(f'\n{fout_nm} was saved.\n')
